/* -*- Mode: C; indent-tabs-mode: t; c-basic-offset: 8; tab-width: 8 -*-  */
/* spfquery.c
 *
 * Copyright (C) 2019  Giam Teck Choon <choon@choon.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301, USA.
 */

#include <string.h>
#include <errno.h>
#include <bglibs/iobuf.h>
#include <bglibs/str.h>
#include <bglibs/dict.h>
#include <netinet/in.h> /* inet_ functions / structs for spf */
#include <spf2/spf.h>
#include <getopt.h>
#include "choonvdeliver.h"

const char conf_qmail[] = "/var/qmail";
static SPF_server_t *spf_server = NULL;
static SPF_request_t *spf_request = NULL;
static SPF_response_t *spf_response = NULL;
static SPF_response_t *spf_response_2mx = NULL;

static str tmpstr = {0,0,0};
static str myhostname = {0,0,0};
static str ipstr = {0,0,0};
static str helostr = {0,0,0};
static str rcptostr = {0,0,0};
static str mailfromstr = {0,0,0};
static int debug = 0;

static void errmsgf(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	obuf_putfv(&errbuf, format, ap);
	obuf_flush(&errbuf);
	va_end(ap);
}

static void errmsg(const char *msg)
{
	obuf_putsflush(&errbuf, msg);
}

static void outmsgf(const char *format, ...)
{
	va_list ap;
	va_start(ap, format);
	obuf_putfv(&outbuf, format, ap);
	obuf_flush(&outbuf);
	va_end(ap);
}

static void outmsg(const char *msg)
{
	obuf_putsflush(&outbuf, msg);
}

static void myexit(int code)
{
	if (tmpstr.s) str_free(&tmpstr);
	if (ipstr.s) str_free(&ipstr);
	if (helostr.s) str_free(&helostr);
	if (rcptostr.s) str_free(&rcptostr);
	if (mailfromstr.s) str_free(&mailfromstr);
	if (myhostname.s) str_free(&myhostname);
	if (spf_request)
		SPF_request_free(spf_request);
	if (spf_server)
		SPF_server_free(spf_server);
	if (spf_response)
		SPF_response_free(spf_response);
	if (spf_response_2mx)
		SPF_response_free(spf_response_2mx);
	if (code) exit(code);
}

static int str_getleft(str *i, str *o, const char sep)
{
	int x;
	x = str_findlast(i, sep);
	if (x < 0) return 0;
	if (o->s || o->len > 0) str_truncate(o, 0);
	if (str_copyb(o, i->s, x)) return 1;
	return 0;
}

static int str_getright(str *i, str *o, const char sep)
{
	int x;
	x = str_findlast(i, sep);
	if (x < 0) return 0;
	x++;
	if (o->s || o->len > 0) str_truncate(o, 0);
	if (str_copyb(o, i->s+x, i->len-x)) return 1;
	return 0;
}

static void showusage(void)
{
	errmsg("Usage:\n"
	       "\n"
	       "spfquery options ...\n"
	       "\n"
	       "Valid options are:\n"
	       "--ip    <IP address>     The IP address that is sending email\n"
	       "--helo  <domain name>    The domain name given on the SMTP HELO\n"
	       "                         command.  This is only needed if the\n"
	       "                         -sender option is not given.\n"
	       "--sender <email address>  The email address used as the\n"
	       "                         envelope-from.  If no username (local\n"
	       "                         part) is given, 'postmaster' will be\n"
	       "                         assumed.\n"
	       "--rcpt <email addresses> A comma separated lists of email addresses\n"
	       "                         that will have email from their secondary\n"
	       "                         MXes automatically allowed.\n"
	       "\n"
	       "The --helo and --rcpt are optional.\n"
	       );
	myexit(1);
}

static void cli_arguments(int argc, char *argv[])
{
	static struct option long_options[] =
	{
		{"ip",      required_argument, NULL,  'i'},
		{"sender",  required_argument, NULL,  's'},
		{"helo",    optional_argument, NULL,  'h'},
		{"rcpt",    optional_argument, NULL,  'r'},
		{"verbose", no_argument,       NULL,  'v'},
		{0,         0,                 0,       0}
	};
	int long_index = 0;
	int opt = 0;
	while ((opt = getopt_long(argc, argv, "vrhi:s:",
				  long_options, &long_index)) != -1) {
		switch (opt)
		{
			case 'i':
				str_copys(&ipstr, optarg);
				break;
			case 's':
				str_copys(&mailfromstr, optarg);
				break;
			case 'h':
				str_copys(&helostr, optarg);
				break;
			case 'r':
				str_copys(&rcptostr, optarg);
				break;
			case 'v':
				debug = 1;
				break;
			default:
				showusage();
				myexit(1);
				break;
		}
	}
	if (optind < argc) {
		errmsg("non-option ARGV-elements:\n");
		while (optind < argc) {
			errmsg(argv[optind++]);
		}
		errmsg("\n");
		myexit(1);
	}
	if (!ipstr.s) {
		errmsg("--ip is required!\n");
		myexit(1);
	}
	if (!mailfromstr.s) {
		errmsg("--sender is required!\n");
		myexit(1);
	}
}

int main(int argc, char *argv[])
{

	cli_arguments(argc, argv);

	//CATCH_NE_ZERO(str_copys(&mailfromstr, argv[1]), myexit(1), debug);
	//CATCH_NE_ONE(str_copys(&mailfromstr, argv[1]), myexit(1), debug);
	outmsg("Sender Email Address: ");
	if (mailfromstr.s)
		outmsg(mailfromstr.s);
	else
		outmsg("Required!\n");
	outmsg("\n");
	//CATCH_NE_ONE(str_copys(&ipstr, argv[2]), myexit(1), debug);
	outmsg("Sender IP Address: ");
	if (ipstr.s)
		outmsg(ipstr.s);
	else
		outmsg("Required!\n");
	outmsg("\n");
	//CATCH_NE_ONE(str_copys(&helostr, argv[3]), myexit(1), debug);
	if (helostr.s) {
		outmsg("Sender HELO: ");
		outmsg(helostr.s);
		outmsg("\n");
	}
	if (rcptostr.s) {
		//CATCH_NE_ONE(str_copys(&rcptostr, argv[4]), myexit(1), debug);
		outmsg("RCPTO Addresses: ");
		outmsg(rcptostr.s);
		outmsg("\n");
	}
	int i = str_findlast(&mailfromstr, '@');
	if (i >= 0) {
		CATCH_NE_ONE(str_getleft(&mailfromstr, &tmpstr, '@'),
			     myexit(1), debug);
		outmsgf("{email user='}s{'\n}", tmpstr.s);
		CATCH_NE_ONE(str_getright(&mailfromstr, &tmpstr, '@'),
			     myexit(1), debug);
		outmsgf("{domain='}s{'\n}", tmpstr.s);
	}

	/* https://github.com/shevek/libspf2/blob/master/src/spf_example/spf_example.c */
	/*
	 * set up the SPF server
	 *
	 * Configurations contain malloc'd data so must be
	 * destroyed when you are finished.
	 */
	CATCH_NULL(spf_server = SPF_server_new(SPF_DNS_CACHE, 0), myexit(1), debug);
	/*
	 * Create a new request.
	 *
	 * The SPF request contains all the data needed to process
	 * the SPF check. Requests are malloc'd so it must be
	 * destroyed when you are finished with it.
	 */
	spf_request = SPF_request_new(spf_server);

	if(strchr(ipstr.s, ':')) {
		if ( SPF_request_set_ipv6_str( spf_request, ipstr.s ) ) {
			errmsg("Invalid IPv6 address.\n");
			myexit(255);
		}
	} else {
		if ( SPF_request_set_ipv4_str( spf_request, ipstr.s ) ) {
			errmsg("Invalid IPv4 address.\n");
			myexit(255);
		}
	}

	/*
	 * record the HELO domain name of the client (sending) MTA from
	 * the SMTP HELO or EHLO commands
	 *
	 * This domain name will be used if the envelope from address is
	 * null (e.g. MAIL FROM:<>).  This happens when a bounce is being
	 * sent and, in effect, it is the client MTA that is sending the
	 * message.
	 */
	if (helostr.s && SPF_request_set_helo_dom( spf_request, helostr.s ) ) {
		errmsg("Invalid HELO domain.\n");
		myexit(255);
	}
	/*
	 * record the envelope from email address from the SMTP MAIL FROM:
	 * command.
	 */

	if ( SPF_request_set_env_from( spf_request, mailfromstr.s ) ) {
		errmsg("Invalid envelope from address.\n");
		myexit(255);
	}
	
	/*
	 * now that we have all the information, see what the result of
	 * the SPF check is.
	 */

	SPF_request_query_mailfrom(spf_request, &spf_response);

	/*
	 * If the sender MAIL FROM check failed, then for each SMTP RCPT TO
	 * command, the mail might have come from a secondary MX for that
	 * domain.
	 *
	 * Note that most MTAs will also check the RCPT TO command to make sure
	 * that it is ok to accept. This SPF check won't give a free pass
	 * to all secondary MXes from all domains, just the one specified by
	 * the rcpt_to address. It is assumed that the MTA checks (at some
	 * point) that we are also a valid primary or secondary for the domain.
	 */
	if (SPF_response_result(spf_response) != SPF_RESULT_PASS && rcptostr.s) {
		SPF_request_query_rcptto(spf_request, &spf_response_2mx, rcptostr.s);
		/*
		 * We might now have a PASS if the mail came from a client which
		 * is a secondary MX from the domain specified in opt_rcpt_to.
		 *
		 * If not, then the RCPT TO: address must have been a domain for
		 * which the client is not a secondary MX, AND the MAIL FROM: domain
		 * doesn't doesn't return 'pass' from SPF_result()
		 */
		if (SPF_response_result(spf_response_2mx) == SPF_RESULT_PASS) {
			errmsg("pass from secondary MX via ");
			errmsg(rcptostr.s);
			errmsg("\n");
		}
	}

	 /*
	 * If the result is something like 'neutral', you probably
	 * want to accept the email anyway, just like you would
	 * when SPF_result() returns 'neutral'.
	 *
	 * It is possible that you will completely ignore the results
	 * until the SMPT DATA command.
	 */
	outmsgf("{result code='}d{' result message='}s{'\n}",
	        SPF_response_result(spf_response),
	        SPF_strresult(SPF_response_result(spf_response)));
	outmsgf("{error code='}d{' error message='}s{'\n}",
	        SPF_response_errcode(spf_response),
	        SPF_strerror(SPF_response_errcode(spf_response)));
	for (i = 0; i < SPF_response_messages(spf_response); i++) {
		SPF_error_t *err = SPF_response_message(spf_response, i);
		outmsgf("s{_msg = (}d{) }s{\n}",
		        (SPF_error_errorp(err) ? "warn" : "err"),
		        SPF_error_code(err),
		        SPF_error_message(err));
	}

	outmsg("\n");
	outmsg("SPF result:\n");
	outmsg(SPF_strresult( SPF_response_result(spf_response) ));
	outmsg("\n");
	outmsg("SMTP comment:\n");

#define VALID_STR(x) (x ? x : "")

	outmsg(VALID_STR(SPF_response_get_smtp_comment(spf_response)));
	outmsg("\n");
	outmsg("Header comment:\n");
	outmsg(VALID_STR(SPF_response_get_header_comment(spf_response)));
	outmsg("\n");
	outmsg("Received SPF:\n");
	outmsg(VALID_STR(SPF_response_get_received_spf(spf_response)));
	outmsg("\n");

#undef VALID_STR
	//static dict dmnp;
	//if (!dict_load_list(&dmnp, "nodmarcpolicy", 0, 0)) {
	//	errmsg("nodmarcpolicy file not found!\n");
	//}
	
	myexit(0);
	return 0;
}
